# -*- coding: utf-8 -*-

"""
/***************************************************************************
 ReservoirAndBasinAnalysis
                                 A QGIS plugin
 This plugin offers some analysis tools for reservoirs.
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2024-07-13
        copyright            : (C) 2024 by João Vitor Pimenta
        email                : jvpjoaopimenta@gmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

__author__ = 'João Vitor Pimenta'
__date__ = '2024-07-13'
__copyright__ = '(C) 2024 by João Vitor Pimenta'

# This will get replaced with a git SHA1 when you do a git archive

__revision__ = '$Format:%H$'

from qgis.PyQt.QtCore import QCoreApplication
from qgis.PyQt.QtGui import QIcon
from qgis.core import (QgsFeatureSink,
                       QgsProcessingAlgorithm,
                       QgsProcessingParameterRasterLayer,
                       QgsProcessingException,
                       QgsProcessingParameterVectorDestination,
                       QgsProcessingParameterPoint,
                       QgsProcessingParameterNumber,
                       QgsProcessingParameterEnum,
                       QgsProcessingParameterFeatureSource,
                       QgsProcessing)
from .algorithms.algorithmFloodedArea import executePluginForArea, executePluginForCoord
import os


class createFloodedAreaAlgorithm(QgsProcessingAlgorithm):
    """
    This is an example algorithm that takes a vector layer and
    creates a new identical one.

    It is meant to be used as an example of how to create your own
    algorithms and explain methods and variables used to do it. An
    algorithm like this will be available in all elements, and there
    is not need for additional work.

    All Processing algorithms should extend the QgsProcessingAlgorithm
    class.
    """

    # Constants used to refer to parameters and outputs. They will be
    # used when calling the algorithm from another algorithm, or when
    # calling from the QGIS console.

    INPUT_DEM = 'INPUT_DEM'
    DRAINAGE_AREA = 'DRAINAGE_AREA'
    INPUT_COORDINATES = 'INPUT_COORDINATES'
    INPUT_PARAMETER = 'INPUT_PARAMETER'
    HEIGHT_PARAMETER = 'HEIGHT (m)'
    ELEVATION_PARAMETER = 'ELEVATION (m)'
    AREA_PARAMETER = 'AREA (m2)'
    VOLUME_PARAMETER = 'VOLUME (m3)'
    VERTICAL_SPACING = 'VERTICAL SPACING (m)'
    FLOODED_AREA = 'FLOODED_AREA'

    def initAlgorithm(self, config):
        """
        Here we define the inputs and output of the algorithm, along
        with some other properties.
        """

        # We add the input vector features source. It can have any kind of
        # geometry.
        self.addParameter(
            QgsProcessingParameterRasterLayer(
                self.INPUT_DEM,
                self.tr('DEM'),
            )
        )

        self.addParameter(
            QgsProcessingParameterFeatureSource(
                self.DRAINAGE_AREA,
                self.tr('Drainage Area'),
                defaultValue=None,
                optional=True,
                types = [QgsProcessing.TypeVectorPolygon]
            )
        )

        self.addParameter(
            QgsProcessingParameterPoint(
                self.INPUT_COORDINATES,
                'Outlet point coordinates (format: x,y)',
                defaultValue=None,
                optional = True
            )
        )

        self.addParameter(
            QgsProcessingParameterEnum(
                'SELECT_OPTION',
                'Select an parameter to calculate the flooded area',
                options=[
                        self.HEIGHT_PARAMETER,
                        self.ELEVATION_PARAMETER,
                        self.AREA_PARAMETER,
                        self.VOLUME_PARAMETER
                        ],
                defaultValue=0
            )
        )

        self.addParameter(
            QgsProcessingParameterNumber(
                self.INPUT_PARAMETER,
                'Parameter value',
                type=QgsProcessingParameterNumber.Double,
                defaultValue='',
                maxValue=1e12
            )
        )

        self.addParameter(
            QgsProcessingParameterNumber(
                self.VERTICAL_SPACING,
                'Vertical step (in meters):',
                type=QgsProcessingParameterNumber.Double,
                defaultValue='0.00'
            )
        )

        # We add a feature sink in which to store our processed features (this
        # usually takes the form of a newly created vector layer when the
        # algorithm is run in QGIS).

        self.addParameter(
            QgsProcessingParameterVectorDestination(
                self.FLOODED_AREA,
                self.tr('Flooded area')
            )
        )

    def processAlgorithm(self, parameters, context, feedback):
        """
        Here is where the processing itself takes place.
        """
        # Retrieve the feature source and sink. The 'dest_id' variable is used
        # to uniquely identify the feature sink, and must be included in the
        # dictionary returned by the processAlgorithm function.
        demLayer = self.parameterAsRasterLayer(
                                             parameters,
                                             self.INPUT_DEM,
                                             context
                                             )
        coordinates = self.parameterAsPoint(
                                            parameters,
                                            self.INPUT_COORDINATES,
                                            context
                                            )
        parameterValue = self.parameterAsDouble(
                                                parameters,
                                                self.INPUT_PARAMETER,
                                                context
                                                )
        parameterNumber = self.parameterAsEnum(
                                               parameters,
                                               'SELECT_OPTION',
                                               context
                                               )
        selectedParameter = [
                             self.HEIGHT_PARAMETER,
                             self.ELEVATION_PARAMETER,
                             self.AREA_PARAMETER,
                             self.VOLUME_PARAMETER,
                             ][parameterNumber]
        drainageAreaInput = self.parameterAsVectorLayer(
                                                        parameters,
                                                        self.DRAINAGE_AREA,
                                                        context
                                                        )
        verticalSpacingInput = self.parameterAsDouble(
                                                        parameters,
                                                        self.VERTICAL_SPACING,
                                                        context
                                                        )
        if drainageAreaInput is None and coordinates.isEmpty():
            raise QgsProcessingException(
                'Outlet point or drainage area, must be provided!'
                                         )
        if drainageAreaInput is not None and not coordinates.isEmpty():
            raise QgsProcessingException(
                'Only one, outlet point or drainage area, must be provided!'
                                         )
        if verticalSpacingInput <0:
            raise QgsProcessingException(
                'Vertical spacing needs be positive!'
                                         )
        demLayerExt = demLayer.extent()


        # Compute the number of steps to display within the progress bar and
        # get features from source

        if drainageAreaInput is not None:
            if demLayerExt.contains(drainageAreaInput.extent()) is False:
                raise QgsProcessingException(
                "The drainage area must be in the DEM extension!"
                )

            floodedArea = executePluginForArea(demLayer,
                                        drainageAreaInput,
                                        selectedParameter,
                                        parameterValue,
                                        verticalSpacingInput)
        if not coordinates.isEmpty():
            if demLayerExt.contains(coordinates) is False:
                raise QgsProcessingException(
                "The point coordinates must be in the DEM extension!"
                )
            x = coordinates.x()
            y = coordinates.y()
            floodedArea = executePluginForCoord(demLayer,
                                        selectedParameter,
                                        parameterValue,
                                        x,y,
                                        verticalSpacingInput)

        (FA, dest_idb) = self.parameterAsSink(parameters,
                                              self.FLOODED_AREA,
                                              context,
                                              floodedArea.fields(),
                                              floodedArea.wkbType(),
                                              floodedArea.sourceCrs(),
                                              layerOptions=["ENCODING=UTF-8"])

        if floodedArea.featureCount():
            total = 100.0 / floodedArea.featureCount()
        else:
            total = 0

        featuresFA = floodedArea.getFeatures()

        for current, feature in enumerate(featuresFA):
            # Stop the algorithm if cancel button has been clicked
            if feedback.isCanceled():
                break

            # Add a feature in the sink
            FA.addFeature(feature, QgsFeatureSink.FastInsert)

            # Update the progress bar
            feedback.setProgress(int(current * total))


        return {self.FLOODED_AREA:dest_idb}



    def name(self):
        """
        Returns the algorithm name, used for identifying the algorithm. This
        string should be fixed for the algorithm, and must not be localised.
        The name should be unique within each provider. Names should contain
        lowercase alphanumeric characters only and no spaces or other
        formatting characters.
        """
        return 'Create a flooded area'

    def displayName(self):
        """
        Returns the translated algorithm name, which should be used for any
        user-visible display of the algorithm name.
        """
        return self.tr(self.name())

    def group(self):
        """
        Returns the name of the group this algorithm belongs to. This string
        should be localised.
        """
        return self.tr(self.groupId())

    def groupId(self):
        """
        Returns the unique ID of the group this algorithm belongs to. This
        string should be fixed for the algorithm, and must not be localised.
        The group id should be unique within each provider. Group id should
        contain lowercase alphanumeric characters only and no spaces or other
        formatting characters.
        """
        return 'Reservoir tools'

    def icon(self):
        """
        Should return a QIcon which is used for your provider inside
        the Processing toolbox.
        """
        return QIcon(os.path.join(os.path.dirname(__file__), "icons", "icon.png"))

    def tr(self, string):
        return QCoreApplication.translate('Processing', string)

    def createInstance(self):
        return createFloodedAreaAlgorithm()
